# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Various function for graph rerouting."""

from __future__ import annotations

from . import subgraph as _subgraph
from . import util as _util
from tensorflow.python.framework import ops as _tf_ops

from tensorflow.python.util.all_util import remove_undocumented

_allowed_symbols = [
    "swap_ts",
    "reroute_ts",
    "swap_inputs",
    "reroute_inputs",
    "swap_outputs",
    "reroute_outputs",
    "swap_ios",
    "reroute_ios",
    "remove_control_inputs",
    "add_control_inputs",
]


def _check_ts_compatibility(ts0, ts1):
    """Make sure the shape and dtype of the two tensor's lists are compatible.

    Args:
      ts0: an object convertible to a list of `tf.Tensor`.
      ts1: an object convertible to a list of `tf.Tensor`.
    Raises:
      ValueError: if any pair of tensors (same index in ts0 and ts1) have
        a dtype or a shape which is not compatible.
    """
    ts0 = _util.make_list_of_t(ts0)
    ts1 = _util.make_list_of_t(ts1)
    if len(ts0) != len(ts1):
        raise ValueError("ts0 and ts1 have different sizes: {} != {}".format(len(ts0), len(ts1)))
    for t0, t1 in zip(ts0, ts1):
        # check dtype
        dtype0, dtype1 = t0.dtype, t1.dtype
        if not dtype0.is_compatible_with(dtype1):
            raise ValueError("Dtypes {} and {} are not compatible.".format(dtype0, dtype1))
        # check shape
        shape0, shape1 = t0.get_shape(), t1.get_shape()
        if not shape0.is_compatible_with(shape1):
            raise ValueError("Shapes {} and {} are not compatible.".format(shape0, shape1))


class _RerouteMode:
    """Enums for reroute's mode.

    swap: the end of tensors a and b are swapped.
    a2b:  the end of the tensor a are also rerouted to the end of the tensor b
      (the end of b is left dangling).
    b2a:  the end of the tensor b are also rerouted to the end of the tensor a
      (the end of a is left dangling).
    """

    swap, a2b, b2a = range(3)

    @classmethod
    def check(cls, mode):
        """Check swap mode.

        Args:
          mode: an integer representing one of the modes.
        Returns:
          A tuple `(a2b, b2a)` boolean indicating what rerouting needs doing.
        Raises:
          ValueError: if mode is outside the enum range.
        """
        if mode == cls.swap:
            return True, True
        elif mode == cls.b2a:
            return False, True
        elif mode == cls.a2b:
            return True, False
        else:
            raise ValueError("Unknown _RerouteMode: {}".format(mode))


def _reroute_t(t0, t1, consumers1, can_modify=None, cannot_modify=None):
    """Reroute the end of the tensors (t0,t1).

    Warning: this function is directly manipulating the internals of the
    `tf.Graph`.

    Args:
      t0: a tf.Tensor.
      t1: a tf.Tensor.
      consumers1: The consumers of t1 which needs to be rerouted.
      can_modify: iterable of operations which can be modified. Any operation
        outside within_ops will be left untouched by this function.
      cannot_modify: iterable of operations which cannot be modified.
        Any operation within cannot_modify will be left untouched by this
        function.
    Returns:
      The number of individual modifications made by the function.
    """
    nb_update_inputs = 0
    if can_modify is not None:
        consumers1 &= can_modify
    if cannot_modify is not None:
        consumers1 -= cannot_modify
    consumers1_indices = {}
    for consumer1 in consumers1:
        consumers1_indices[consumer1] = [i for i, t in enumerate(consumer1.inputs) if t is t1]
    for consumer1 in consumers1:
        for i in consumers1_indices[consumer1]:
            consumer1._update_input(i, t0)  # pylint: disable=protected-access
            nb_update_inputs += 1
    return nb_update_inputs


def _reroute_ts(ts0, ts1, mode, can_modify=None, cannot_modify=None):
    """Reroute the end of the tensors in each pair (t0,t1) in ts0 x ts1.

    This function is the back-bone of the Graph-Editor. It is essentially a thin
    wrapper on top of the tf.Operation._update_input.

    Given a pair of tensor t0, t1 in ts0 x ts1, this function re-route the end
    of t0 and t1 in three possible ways:
    1) The reroute mode is "a<->b" or "b<->a": the tensors' end are swapped. After
    this operation, the previous consumers of t0 are now consumers of t1 and
    vice-versa.
    2) The reroute mode is "a->b": the tensors' end of t0 are re-routed to the
    tensors's end of t1 (which are left dangling). After this operation, the
    previous consumers of t0 are still consuming t0 but the previous consumers of
    t1 are not also consuming t0. The tensor t1 has no consumer.
    3) The reroute mode is "b->a": this mode is the symmetric of the "a->b" mode.

    Note that this function is re-routing the end of two tensors, not the start.
    Re-routing the start of two tensors is not supported by this library. The
    reason for that is the following: TensorFlow, by design, creates a strong bond
    between an op and its output tensor. This Graph editor follows this design and
    treats an operation A and its generating tensors {t_i} as an entity which
    cannot be broken. In other words, an op cannot be detached from any of its
    output tensors, ever. But it is possible to detach an op from its input
    tensors, which is what this function concerns itself with.

    Warning: this function is directly manipulating the internals of the tf.Graph.

    Args:
      ts0: an object convertible to a list of `tf.Tensor`.
      ts1: an object convertible to a list of `tf.Tensor`.
      mode: what to do with those tensors: "a->b" or "b<->a" for swaping and
        "a->b" or "b->a" for one direction re-routing.
      can_modify: iterable of operations which can be modified. Any operation
        outside within_ops will be left untouched by this function.
      cannot_modify: iterable of operations which cannot be modified.
        Any operation within cannot_modify will be left untouched by this
        function.
    Returns:
      The number of individual modifications made by the function.
    Raises:
      TypeError: if `ts0` or `ts1` cannot be converted to a list of `tf.Tensor`.
      TypeError: if `can_modify` or `cannot_modify` is not `None` and cannot be
        converted to a list of `tf.Operation`.
    """
    a2b, b2a = _RerouteMode.check(mode)
    ts0 = _util.make_list_of_t(ts0)
    ts1 = _util.make_list_of_t(ts1)
    _check_ts_compatibility(ts0, ts1)
    if cannot_modify is not None:
        cannot_modify = frozenset(_util.make_list_of_op(cannot_modify))
    if can_modify is not None:
        can_modify = frozenset(_util.make_list_of_op(can_modify))
    nb_update_inputs = 0
    precomputed_consumers = []
    # precompute consumers to avoid issue with repeated tensors:
    for t0, t1 in zip(ts0, ts1):
        consumers0 = set(t0.consumers())
        consumers1 = set(t1.consumers())
        precomputed_consumers.append((consumers0, consumers1))
    for t0, t1, consumers in zip(ts0, ts1, precomputed_consumers):
        if t0 is t1:
            continue  # Silently ignore identical tensors.
        consumers0, consumers1 = consumers
        if a2b:
            nb_update_inputs += _reroute_t(t0, t1, consumers1, can_modify, cannot_modify)
        if b2a:
            nb_update_inputs += _reroute_t(t1, t0, consumers0, can_modify, cannot_modify)
    return nb_update_inputs


def swap_ts(ts0, ts1, can_modify=None, cannot_modify=None):
    """For each tensor's pair, swap the end of (t0,t1).

        B0 B1     B0 B1
        |  |    =>  X
        A0 A1     A0 A1

    Args:
      ts0: an object convertible to a list of `tf.Tensor`.
      ts1: an object convertible to a list of `tf.Tensor`.
      can_modify: iterable of operations which can be modified. Any operation
        outside within_ops will be left untouched by this function.
      cannot_modify: iterable of operations which cannot be modified.
        Any operation within cannot_modify will be left untouched by this
        function.
    Returns:
      The number of individual modifications made by the function.
    Raises:
      TypeError: if ts0 or ts1 cannot be converted to a list of tf.Tensor.
      TypeError: if can_modify or cannot_modify is not None and cannot be
        converted to a list of tf.Operation.
    """
    return _reroute_ts(ts0, ts1, _RerouteMode.swap, can_modify, cannot_modify)


def reroute_ts(ts0, ts1, can_modify=None, cannot_modify=None):
    """For each tensor's pair, replace the end of t1 by the end of t0.

        B0 B1     B0 B1
        |  |    => |/
        A0 A1     A0 A1

    The end of the tensors in ts1 are left dangling.

    Args:
      ts0: an object convertible to a list of `tf.Tensor`.
      ts1: an object convertible to a list of `tf.Tensor`.
      can_modify: iterable of operations which can be modified. Any operation
        outside within_ops will be left untouched by this function.
      cannot_modify: iterable of operations which cannot be modified. Any
        operation within cannot_modify will be left untouched by this function.
    Returns:
      The number of individual modifications made by the function.
    Raises:
      TypeError: if ts0 or ts1 cannot be converted to a list of tf.Tensor.
      TypeError: if can_modify or cannot_modify is not None and cannot be
        converted to a list of tf.Operation.
    """
    return _reroute_ts(ts0, ts1, _RerouteMode.a2b, can_modify, cannot_modify)


def _reroute_sgv_remap(sgv0, sgv1, mode):
    """Remap in place the inputs of two subgraph views to mimic the reroute.

    This function is meant to used by reroute_inputs only.

    Args:
      sgv0: the first subgraph to have its inputs remapped.
      sgv1: the second subgraph to have its inputs remapped.
      mode: reroute mode, see _reroute_ts(...).
    Raises:
      TypeError: if svg0 or svg1 are not SubGraphView.
      ValueError: if sgv0 and sgv1 do not belong to the same graph.
    """
    a2b, b2a = _RerouteMode.check(mode)
    if not isinstance(sgv0, _subgraph.SubGraphView):
        raise TypeError("Expected a SubGraphView, got {}".format(type(sgv0)))
    if not isinstance(sgv1, _subgraph.SubGraphView):
        raise TypeError("Expected a SubGraphView, got {}".format(type(sgv1)))
    _util.check_graphs(sgv0, sgv1)
    sgv0_ = sgv0.copy()
    sgv1_ = sgv1.copy()
    # pylint: disable=protected-access
    if a2b and b2a:
        (sgv0_._input_ts, sgv1_._input_ts) = (sgv1_._input_ts, sgv0_._input_ts)
        (sgv0_._passthrough_ts, sgv1_._passthrough_ts) = (sgv1_._passthrough_ts, sgv0_._passthrough_ts)
    elif a2b:
        sgv1_._input_ts = sgv0_._input_ts[:]
        sgv1_._passthrough_ts = sgv0_._passthrough_ts[:]
    elif b2a:
        sgv0_._input_ts = sgv1_._input_ts[:]
        sgv0_._passthrough_ts = sgv1_._passthrough_ts[:]
    # pylint: enable=protected-access

    # Update the passthrough outputs as well.
    def update_passthrough_outputs(a, b):
        # pylint: disable=protected-access
        for i, t in enumerate(b._output_ts):
            if t in a._passthrough_ts:
                ii = a._input_ts.index(t)
                b._output_ts[i] = b._input_ts[ii]
        # pylint: enable=protected-access

    if a2b:
        update_passthrough_outputs(sgv0_, sgv1_)
    if b2a:
        update_passthrough_outputs(sgv1_, sgv0_)

    # in-place
    # pylint: disable=protected-access
    sgv0._assign_from(sgv0_)
    sgv1._assign_from(sgv1_)
    # pylint: enable=protected-access


def _reroute_sgv_inputs(sgv0, sgv1, mode):
    """Re-route all the inputs of two subgraphs.

    Args:
      sgv0: the first subgraph to have its inputs swapped. This argument is
        converted to a subgraph using the same rules than the function
        subgraph.make_view.
      sgv1: the second subgraph to have its inputs swapped. This argument is
        converted to a subgraph using the same rules than the function
        subgraph.make_view.
      mode: reroute mode, see _reroute_ts(...).
    Returns:
      A tuple `(sgv0, sgv1)` of subgraph views with their inputs swapped.
        Note that the function argument sgv0 and sgv1 are also modified in place.
    Raises:
      StandardError: if sgv0 or sgv1 cannot be converted to a SubGraphView using
        the same rules than the function subgraph.make_view.
    """
    sgv0 = _subgraph.make_view(sgv0)
    sgv1 = _subgraph.make_view(sgv1)
    _util.check_graphs(sgv0, sgv1)
    can_modify = sgv0.ops + sgv1.ops
    # also allow consumers of passthrough to be modified:
    can_modify += _util.get_consuming_ops(sgv0.passthroughs)
    can_modify += _util.get_consuming_ops(sgv1.passthroughs)
    _reroute_ts(sgv0.inputs, sgv1.inputs, mode, can_modify=can_modify)
    _reroute_sgv_remap(sgv0, sgv1, mode)
    return sgv0, sgv1


def _reroute_sgv_outputs(sgv0, sgv1, mode):
    """Re-route all the outputs of two operations.

    Args:
      sgv0: the first subgraph to have its outputs swapped. This argument is
        converted to a subgraph using the same rules than the function
        subgraph.make_view.
      sgv1: the second subgraph to have its outputs swapped. This argument is
        converted to a subgraph using the same rules than the function
        subgraph.make_view.
      mode: reroute mode, see _reroute_ts(...).
    Returns:
      A tuple `(sgv0, sgv1)` of subgraph views with their outputs swapped.
        Note that the function argument sgv0 and sgv1 are also modified in place.
    Raises:
      StandardError: if sgv0 or sgv1 cannot be converted to a SubGraphView using
        the same rules than the function subgraph.make_view.
    """
    sgv0 = _subgraph.make_view(sgv0)
    sgv1 = _subgraph.make_view(sgv1)
    _util.check_graphs(sgv0, sgv1)
    cannot_modify = sgv0.ops + sgv1.ops
    _reroute_ts(sgv0.outputs, sgv1.outputs, mode, cannot_modify=cannot_modify)
    return sgv0, sgv1


def _reroute_sgv(sgv0, sgv1, mode):
    """Re-route both the inputs and the outputs of the two subgraph views.

    This involves swapping all the inputs/outputs of the two subgraph views.

    Args:
      sgv0: the first subgraph to be swapped. This argument is converted to a
        subgraph using the same rules than the function subgraph.make_view.
      sgv1: the second subgraph to be swapped. This argument is converted to a
        subgraph using the same rules than the function subgraph.make_view.
      mode: reroute mode, see _reroute_ts(...).
    Returns:
      A tuple `(sgv0, sgv1)` of subgraph views with their outputs and inputs
        swapped.
        Note that the function argument sgv0 and sgv1 are also modified in place.
    Raises:
      StandardError: if sgv0 or sgv1 cannot be converted to a SubGraphView using
        the same rules than the function subgraph.make_view.
    """
    _reroute_sgv_outputs(sgv0, sgv1, mode)
    _reroute_sgv_inputs(sgv0, sgv1, mode)
    return sgv0, sgv1


def swap_inputs(sgv0, sgv1):
    """Swap all the inputs of sgv0 and sgv1 (see reroute_inputs)."""
    return _reroute_sgv_inputs(sgv0, sgv1, _RerouteMode.swap)


def reroute_inputs(sgv0, sgv1):
    """Re-route all the inputs of two subgraphs.

    Args:
      sgv0: the first subgraph to have its inputs swapped. This argument is
        converted to a subgraph using the same rules than the function
        subgraph.make_view.
      sgv1: the second subgraph to have its inputs swapped. This argument is
        converted to a subgraph using the same rules than the function
        subgraph.make_view.
    Returns:
      A tuple `(sgv0, sgv1)` of subgraph views with their inputs swapped.
        Note that the function argument sgv0 and sgv1 are also modified in place.
    Raises:
      StandardError: if sgv0 or sgv1 cannot be converted to a SubGraphView using
        the same rules than the function subgraph.make_view.
    """
    return _reroute_sgv_inputs(sgv0, sgv1, _RerouteMode.a2b)


def swap_outputs(sgv0, sgv1):
    """Swap all the outputs of sgv0 and sgv1 (see reroute_outputs)."""
    return _reroute_sgv_outputs(sgv0, sgv1, _RerouteMode.swap)


def reroute_outputs(sgv0, sgv1):
    """Re-route all the outputs of two operations.

    Args:
      sgv0: the first subgraph to have its outputs swapped. This argument is
        converted to a subgraph using the same rules than the function
        subgraph.make_view.
      sgv1: the second subgraph to have its outputs swapped. This argument is
        converted to a subgraph using the same rules than the function
        subgraph.make_view.
    Returns:
      A tuple `(sgv0, sgv1)` of subgraph views with their outputs swapped.
        Note that the function argument sgv0 and sgv1 are also modified in place.
    Raises:
      StandardError: if sgv0 or sgv1 cannot be converted to a SubGraphView using
        the same rules than the function subgraph.make_view.
    """
    return _reroute_sgv_outputs(sgv0, sgv1, _RerouteMode.a2b)


def swap_ios(sgv0, sgv1):
    """Swap the inputs and outputs of sgv1 to sgv0 (see _reroute_sgv)."""
    return _reroute_sgv(sgv0, sgv1, _RerouteMode.swap)


def reroute_ios(sgv0, sgv1):
    """Re-route the inputs and outputs of sgv0 to sgv1 (see _reroute_sgv)."""
    return _reroute_sgv(sgv0, sgv1, _RerouteMode.a2b)


def remove_control_inputs(op, cops):
    """Remove the control inputs cops from co.

    Warning: this function is directly manipulating the internals of the
    `tf.Graph`.

    Args:
      op: a `tf.Operation` from which to remove the control inputs.
      cops: an object convertible to a list of `tf.Operation`.
    Raises:
      TypeError: if op is not a `tf.Operation`.
      ValueError: if any cop in cops is not a control input of op.
    """
    if not isinstance(op, _tf_ops.Operation):
        raise TypeError("Expected a tf.Operation, got: {}", type(op))
    cops = _util.make_list_of_op(cops, allow_graph=False)
    for cop in cops:
        if cop not in op.control_inputs:
            raise ValueError("{} is not a control_input of {}".format(op.name, cop.name))
    control_inputs = [cop for cop in op.control_inputs if cop not in cops]
    # pylint: disable=protected-access
    op._remove_all_control_inputs()
    op._add_control_inputs(control_inputs)
    # pylint: enable=protected-access


def add_control_inputs(op, cops):
    """Add the control inputs cops to op.

    Warning: this function is directly manipulating the internals of the tf.Graph.

    Args:
      op: a tf.Operation to which the control inputs are added.
      cops: an object convertible to a list of `tf.Operation`.
    Raises:
      TypeError: if op is not a tf.Operation
      ValueError: if any cop in cops is already a control input of op.
    """
    if not isinstance(op, _tf_ops.Operation):
        raise TypeError("Expected a tf.Operation, got: {}", type(op))
    cops = _util.make_list_of_op(cops, allow_graph=False)
    for cop in cops:
        if cop in op.control_inputs:
            raise ValueError("{} is already a control_input of {}".format(cop.name, op.name))
    op._add_control_inputs(cops)  # pylint: disable=protected-access


remove_undocumented(__name__, _allowed_symbols)



from __future__ import annotations
import _setup_test_env  # noqa
from returnn.pretrain import pretrain_from_config
from returnn.config import Config


config_dict = {
    "pretrain": "default",
    "num_inputs": 40,
    "num_outputs": 4498,
}


net_dict = {
    "hidden_0": {"class": "linear", "n_out": 7, "dropout": 0.1, "activation": "relu"},
    "hidden_1": {"class": "linear", "n_out": 8, "dropout": 0.1, "activation": "relu", "from": ["hidden_0"]},
    "output": {"class": "softmax", "loss": "ce", "from": ["hidden_1"]},
}


net_dict2 = {
    "lstm0_fw": {"class": "lstm_opt", "n_out": 500, "dropout": 0.1, "sampling": 1, "reverse": False},
    "lstm0_bw": {"class": "lstm_opt", "n_out": 500, "dropout": 0.1, "sampling": 1, "reverse": True},
    "lstm1_fw": {
        "class": "lstm_opt",
        "n_out": 500,
        "dropout": 0.1,
        "sampling": 1,
        "reverse": False,
        "from": ["lstm0_fw", "lstm0_bw"],
    },
    "lstm1_bw": {
        "class": "lstm_opt",
        "n_out": 500,
        "dropout": 0.1,
        "sampling": 1,
        "reverse": True,
        "from": ["lstm0_fw", "lstm0_bw"],
    },
    "lstm2_fw": {
        "class": "lstm_opt",
        "n_out": 500,
        "dropout": 0.1,
        "sampling": 1,
        "reverse": False,
        "from": ["lstm1_fw", "lstm1_bw"],
    },
    "lstm2_bw": {
        "class": "lstm_opt",
        "n_out": 500,
        "dropout": 0.1,
        "sampling": 1,
        "reverse": True,
        "from": ["lstm1_fw", "lstm1_bw"],
    },
    "output": {"class": "softmax", "loss": "ce", "from": ["lstm2_fw", "lstm2_bw"]},
}


def test_config_net_dict1():
    config = Config()
    config.update(config_dict)
    config.typed_dict["network"] = net_dict
    pretrain = pretrain_from_config(config)
    assert pretrain.get_train_num_epochs() == 2
    net1_json = pretrain.get_network_json_for_epoch(1)
    net2_json = pretrain.get_network_json_for_epoch(2)
    net3_json = pretrain.get_network_json_for_epoch(3)
    assert "hidden_0" in net1_json
    assert "hidden_1" not in net1_json
    assert "hidden_0" in net2_json
    assert "hidden_1" in net2_json
    assert net2_json == net3_json


def test_config_net_dict2():
    config = Config()
    config.update(config_dict)
    config.typed_dict["network"] = net_dict2
    pretrain = pretrain_from_config(config)
    assert pretrain.get_train_num_epochs() == 3



# -*- coding: utf8 -*-

from __future__ import annotations

import _setup_test_env  # noqa
import unittest
from returnn.datasets.generating import *
from returnn.datasets.basic import DatasetSeq
from returnn.util.basic import PY3, unicode
import os
import unittest

from returnn.util import better_exchook
from returnn.log import log


my_dir = os.path.dirname(os.path.realpath(__file__))


def test_init():
    dataset = DummyDataset(input_dim=2, output_dim=3, num_seqs=4)
    assert dataset.num_inputs == 2
    assert dataset.num_outputs == {"classes": (3, 1), "data": (2, 2)}
    assert dataset.num_seqs == 4


def test_load_seqs():
    dataset = DummyDataset(input_dim=2, output_dim=3, num_seqs=4)
    dataset.init_seq_order(epoch=1)
    dataset.load_seqs(0, 1)
    dataset.load_seqs(1, 3)


@unittest.skipIf(not os.path.exists("/tmp/enwik8.zip"), "we will not trigger the download")
def test_Enwik8Corpus_batch_num_seqs():
    dataset = Enwik8Corpus(path="/tmp", subset="validation", seq_len=13)
    dataset.init_seq_order(epoch=17)
    data = b""
    n = 0
    while dataset.is_less_than_num_seqs(n) and n < 100:
        dataset.load_seqs(n, n + 1)
        data += bytes(dataset.get_data(n, "data"))
        n += 1

    batch_size = 23
    batch_data = [b"" for i in range(batch_size)]
    dataset = Enwik8Corpus(path="/tmp", subset="validation", seq_len=9, batch_num_seqs=batch_size)
    dataset.init_seq_order(epoch=31)
    n = 0
    while dataset.is_less_than_num_seqs(n) and n < 100:
        dataset.load_seqs(n, n + 1)
        new_data = bytes(dataset.get_data(n, "data"))
        batch_data[n % batch_size] += new_data
        n += 1
    assert data.startswith(batch_data[0])


def test_StaticDataset_custom_keys():
    dataset = StaticDataset([{"source": numpy.array([1, 2, 3]), "target": numpy.array([3, 4, 5, 6, 7])}])
    dataset.init_seq_order(epoch=1)
    assert dataset.num_seqs == 1
    assert dataset.get_data_keys() == ["source", "target"]
    assert dataset.num_outputs["source"][1] == 1
    assert dataset.num_outputs["target"][1] == 1
    dataset.load_seqs(0, 1)
    assert list(dataset.get_data(0, "source")) == [1, 2, 3]
    assert list(dataset.get_data(0, "target")) == [3, 4, 5, 6, 7]


def test_StaticDataset_custom_keys_with_dims():
    dataset = StaticDataset(
        data=[{"source": numpy.array([1, 2, 3]), "target": numpy.array([3, 4, 5, 6, 7])}],
        output_dim={"source": [5, 1], "target": [10, 1]},
    )
    dataset.init_seq_order(epoch=1)
    assert dataset.num_seqs == 1
    assert dataset.get_data_keys() == ["source", "target"]
    assert dataset.num_outputs["source"][1] == 1
    assert dataset.num_outputs["target"][1] == 1
    dataset.load_seqs(0, 1)
    assert list(dataset.get_data(0, "source")) == [1, 2, 3]
    assert list(dataset.get_data(0, "target")) == [3, 4, 5, 6, 7]


def test_StaticDataset_utf8():
    s = "wër"
    print("some unicode str:", s, "repr:", repr(s), "type:", type(s), "len:", len(s))
    assert len(s) == 3
    if PY3:
        assert isinstance(s, str)
        s_byte_list = list(s.encode("utf8"))
    else:
        assert isinstance(s, unicode)
        s_byte_list = list(map(ord, s.encode("utf8")))
    print("utf8 byte list:", s_byte_list)
    assert len(s_byte_list) == 4 > 3
    raw = numpy.array(s_byte_list, dtype="uint8")
    assert raw.tolist() == [119, 195, 171, 114]
    data = StaticDataset([{"data": raw}], output_dim={"data": (255, 1)})
    if "data" not in data.labels:
        data.labels["data"] = [chr(i) for i in range(255)]  # like in SprintDataset
    data.init_seq_order(epoch=1)
    data.load_seqs(0, 1)
    raw_ = data.get_data(seq_idx=0, key="data")
    assert raw.tolist() == raw_.tolist()
    assert data.can_serialize_data(key="data")
    s_serialized = data.serialize_data(key="data", data=raw)
    print("serialized:", s_serialized, "repr:", repr(s_serialized), "type:", type(s_serialized))
    assert s == s_serialized


# might be moved to a separate test_MetaDataset ...
def test_ConcatSeqsDataset():
    num_seqs = 2
    seq_len = 3
    sub_dataset = StaticDataset([{"data": numpy.array(range(1, seq_len + 1))}] * num_seqs)
    from returnn.datasets.meta import ConcatSeqsDataset
    import tempfile

    seq_list_f = tempfile.NamedTemporaryFile(mode="w", prefix="seq-list", suffix=".txt")
    seq_len_f = tempfile.NamedTemporaryFile(mode="w", prefix="seq-len", suffix=".txt")
    with seq_list_f, seq_len_f:
        seq_len_f.write("{\n")
        for i in range(num_seqs):
            seq_list_f.write("seq-%i\n" % i)
            seq_len_f.write("'seq-%i': %i,\n" % (i, seq_len))
        seq_len_f.write("}\n")
        for i in range(0, num_seqs, 2):
            seq_list_f.write("seq-%i;seq-%i\n" % (i, i + 1))
        seq_list_f.flush()
        seq_len_f.flush()
        dataset = ConcatSeqsDataset(dataset=sub_dataset, seq_list_file=seq_list_f.name, seq_len_file=seq_len_f.name)
        dataset.init_seq_order(epoch=1)
        concat_num_seqs = num_seqs + num_seqs // 2
        dataset.load_seqs(0, concat_num_seqs)
        assert dataset.num_seqs == concat_num_seqs == 3
        assert dataset.get_data(0, "data").tolist() == [1, 2, 3]
        assert dataset.get_data(1, "data").tolist() == [1, 2, 3]
        assert dataset.get_data(2, "data").tolist() == [1, 2, 3, 1, 2, 3]


# might be moved to a separate test_MetaDataset ...
def test_ConcatSeqsDataset_repeat_in_between_last_frame_up_to_multiple_of():
    num_seqs = 2
    sub_dataset = StaticDataset([{"data": numpy.array([1, 2])}, {"data": numpy.array([1, 2, 3])}])
    from returnn.datasets.meta import ConcatSeqsDataset
    import tempfile

    seq_list_f = tempfile.NamedTemporaryFile(mode="w", prefix="seq-list", suffix=".txt")
    seq_len_f = tempfile.NamedTemporaryFile(mode="w", prefix="seq-len", suffix=".txt")
    with seq_list_f, seq_len_f:
        seq_len_f.write("%r\n" % ({"seq-0": 2, "seq-1": 3},))
        seq_len_f.flush()
        seq_list_f.write("seq-0\n")
        seq_list_f.write("seq-1\n")
        seq_list_f.write("seq-0;seq-1;seq-1\n")
        seq_list_f.flush()
        concat_num_seqs = 3
        dataset = ConcatSeqsDataset(
            dataset=sub_dataset,
            repeat_in_between_last_frame_up_to_multiple_of={"data": 5},
            seq_list_file=seq_list_f.name,
            seq_len_file=seq_len_f.name,
        )
        dataset.init_seq_order(epoch=1)
        dataset.load_seqs(0, concat_num_seqs)
        assert dataset.num_seqs == concat_num_seqs == 3
        assert dataset.get_data(0, "data").tolist() == [1, 2]
        assert dataset.get_data(1, "data").tolist() == [1, 2, 3]
        assert dataset.get_data(2, "data").tolist() == [1, 2, 2, 2, 2, 1, 2, 3, 3, 3, 1, 2, 3]


def test_BytePairEncoding_unicode():
    bpe = BytePairEncoding(
        bpe_file="%s/bpe-unicode-demo.codes" % my_dir,
        vocab_file="%s/bpe-unicode-demo.vocab" % my_dir,
        unknown_label="<unk>",
    )
    assert bpe.num_labels == 189
    assert bpe.id_to_label(5) == "z"
    assert bpe.label_to_id("z") == 5
    assert bpe.bpe._bpe_codes[("n", "d</w>")] == 1
    assert bpe.id_to_label(6) == "å"
    assert bpe.label_to_id("å") == 6
    assert bpe.bpe._bpe_codes[("à", "nd</w>")] == 2

    def get_bpe_seq(text):
        """
        :param str text:
        :rtype: str
        """
        bpe_label_seq = bpe.get_seq(text)
        res = " ".join(bpe.id_to_label(i) for i in bpe_label_seq)
        print("%r -> %r" % (text, res))
        return res

    assert get_bpe_seq("kod") == "k@@ o@@ d"  # str
    assert get_bpe_seq("kod") == "k@@ o@@ d"  # unicode
    assert get_bpe_seq("råt") == "råt"
    assert (
        get_bpe_seq("råt råt iz ďër iz ďër ám àn iz ďër ë låk ë kod áv dres wër yù wêk dù ďë àsk")
        == "råt råt iz ďër iz ďër ám à@@ n iz ďër ë låk ë k@@ o@@ d áv d@@ r@@ e@@ s w@@ ër yù w@@ ê@@ k dù ďë à@@ s@@ k"
    )


if __name__ == "__main__":
    better_exchook.install()
    if len(sys.argv) <= 1:
        for k, v in sorted(globals().items()):
            if k.startswith("test_"):
                print("-" * 40)
                print("Executing: %s" % k)
                try:
                    v()
                except unittest.SkipTest as exc:
                    print("SkipTest:", exc)
                print("-" * 40)
        print("Finished all tests.")
    else:
        assert len(sys.argv) >= 2
        for arg in sys.argv[1:]:
            print("Executing: %s" % arg)
            if arg in globals():
                globals()[arg]()  # assume function and execute
            else:
                eval(arg)  # assume Python code and execute



"""
Utilities for dimension tags, dimensions, axes.
"""

from __future__ import annotations
from typing import Optional, Union, TypeVar, Sequence, Tuple
from returnn.tensor import Tensor, Dim
import returnn.frontend as rf
from ._backend import get_backend_by_tensor, global_backend

T = TypeVar("T")

__all__ = [
    "range_over_dim",
    "range_over_dim_strided",
    "range_over_merged_dims",
    "replace_dim",
    "set_sparse_dim",
    "dim_match_priority_when_needed",
    "num_elements_of_shape",
    "masked_fraction_of_shape",
]


def range_over_dim(dim: Dim, *, dtype: Optional[str] = None, device: Optional[str] = None) -> Tensor[T]:
    """
    :param dim:
    :param dtype:
    :param device,
    :return: tensor with shape [dim]
    """
    if dim.dyn_size_ext:
        backend = get_backend_by_tensor(dim.dyn_size_ext, fallback=global_backend)
    else:
        backend = global_backend
    return backend.range_over_dim(dim, dtype=dtype, device=device)


def range_over_dim_strided(
    dim: Dim,
    *,
    stride: Union[int, Tensor],
    out_dim: Optional[Dim] = None,
    dtype: Optional[str] = None,
    device: Optional[str] = None,
) -> Tuple[Tensor[T], Dim]:
    """
    :param dim:
    :param stride:
    :param out_dim:
    :param dtype:
    :param device,
    :return: tensor with shape [dim], out_dim
    """
    if out_dim is None:
        out_dim = dim.ceildiv_right(stride)
    return rf.range_over_dim(out_dim, dtype=dtype, device=device) * stride, out_dim


def range_over_merged_dims(
    dims: Sequence[Dim], *, dtype: Optional[str] = None, device: Optional[str] = None
) -> Tensor[T]:
    """
    This is if you want to index into a merged dim.
    Related: :func:`rf.merge_dims`.

    :param dims:
    :param dtype:
    :param device:
    :return: tensor with shape [dim_0, ..., dim_n] -> sparse_dim = merged_dim, where merged_dim = dim_0 * ... * dim_n
    """
    assert len(dims) >= 1
    merged_dim = dims[0]
    for dim in dims[1:]:
        merged_dim *= dim
    indices = rf.range_over_dim(merged_dim, dtype=dtype, device=device)
    if len(dims) > 1:
        indices = rf.split_dims(indices, axis=merged_dim, dims=dims)
    return indices


def replace_dim(source: Tensor, *, in_dim: Dim, out_dim: Optional[Dim] = None) -> Tuple[Tensor, Dim]:
    """
    Also see: :func:`rf.merge_dims`, :func:`rf.split_dims`.

    :param source:
    :param in_dim:
    :param out_dim:
    :return: source with in_dim replaced by out_dim, and new out_dim.
        this does not work for the sparse_dim. see :func:`set_sparse_dim` for that case.
    """
    if not out_dim:
        out_dim = in_dim.copy(same_as_self=False, description="new-dim")
    # noinspection PyProtectedMember
    return source._raw_backend.replace_dim(source, in_dim=in_dim, out_dim=out_dim), out_dim


def set_sparse_dim(source: Tensor, sparse_dim: Dim) -> Tensor:
    """
    :param source:
    :param sparse_dim:
    :return: source with sparse_dim set
    """
    # noinspection PyProtectedMember
    return source._raw_backend.set_sparse_dim(source, sparse_dim)


def dim_match_priority_when_needed(dim: Dim, *other_dims: Dim) -> Dim:
    """
    :return: maybe copy of dim with higher match_priority if needed to distinguish from other_dims

    Why or when is this needed?

    For activation values, this should never be needed,
    and all dims should be unique.

    In case of self-attention, the standard way is to create a separate distinct dim
    to perform the attention reduction over.
    See :class:`SelfAttention`.

    However, in case of weight matrices, it is not unusual to have the same dim for both the input and output,
    so a square weight matrix.
    When reduction is performed in :func:`matmul`, we want to match the input feature dim
    to the dim in the weight matrix with higher match priority.

    So :func:`dim_match_priority_when_needed` would be applied on the input feature dim.

    https://github.com/rwth-i6/returnn/pull/871
    https://github.com/rwth-i6/returnn_common/issues/17#issuecomment-997463222
    """
    if dim in other_dims:
        return dim.copy(match_priority=1)
    return dim


def num_elements_of_shape(dims: Union[Dim, Sequence[Dim]], *, use_mask: bool = True) -> Union[int, Tensor]:
    """
    :param dims:
    :param use_mask:
    :return: num elements of a tensor of shape dims, properly considering masking
    """
    if isinstance(dims, Dim):
        dims = [dims]
    if not use_mask:
        n = 1
        for dim in dims:
            n *= dim.get_dim_value_tensor()
        return n
    if all(dim.is_static() for dim in dims):
        n = 1
        for dim in dims:
            n *= dim.dimension
        return n

    n: Union[int, Tensor] = 1
    postponed_dims = []
    for i, dim in enumerate(dims):
        # E.g. if dim==B, and some other dim dyn_size_ext has B, then we need to postpone this.
        related_dims = []
        for j, dim_ in enumerate(dims):
            if i == j:
                continue
            if dim_.dyn_size_ext and dim in dim_.dyn_size_ext.dims:
                related_dims.append(dim_)
        if not related_dims:
            if dim.is_static():
                n *= dim.dimension
            else:
                n *= dim.dyn_size_ext
        else:
            postponed_dims.append(dim)
    if postponed_dims:
        n: Tensor
        n = rf.reduce_sum(n, axis=postponed_dims)
    return n


def masked_fraction_of_shape(dims: Union[Dim, Sequence[Dim]], *, inverse: bool = False) -> Union[int, float, Tensor]:
    """
    :param dims:
    :param inverse: if True, return the inverse of the fraction
    :return: :func:`num_elements_of_shape`(dims) / prod(dims) if not inverse else prod(dims) / num_elements
    """
    if isinstance(dims, Dim):
        dims = [dims]
    if not any(dim.need_masking() for dim in dims):
        return 1
    num_elems_masked = num_elements_of_shape(dims)
    num_elems_total = 1
    for dim in dims:
        num_elems_total *= dim.get_dim_value_tensor()
    return (num_elems_masked / num_elems_total) if not inverse else (num_elems_total / num_elems_masked)
